/////////////////////////////////////////////////////////////////////////////
// Name:        FiberedFractalTiler.cpp
// Purpose:     Generate TAM tilesets which self-assemble into designed shapes
// Author:      Matthew Patitz <mpatitz@yahoo.com>
// Modified by:
// Created:     Nov/2007
// Copyright:   (c) 2008, Matthew John Patitz
// Licence:     GPL
/////////////////////////////////////////////////////////////////////////////

// ==========================================================================
// headers, declarations, constants
// ==========================================================================
// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wx/statline.h"
#include "wx/wfstream.h"
#include "wx/filedlg.h"
#include "wx/stockitem.h"
#include <wx/fs_mem.h>
#include <wx/fs_arc.h>
#include <wx/cshelp.h>
#include <wx/colordlg.h>

#include "FiberedFractalTiler.h"
#include "fiber.h"

#define FIBERER_ONLY  1


// --------------------------------------------------------------------------
// resources
// --------------------------------------------------------------------------

#if defined(__WXGTK__) || defined(__WXMOTIF__) || defined(__WXMAC__) || defined(__WXMGL__) || defined(__WXX11__)
    // application icon
    #include "mondrian.xpm"

    // bitmap buttons for the toolbar
    #include "bitmaps/reset.xpm"
    #include "bitmaps/open.xpm"
    #include "bitmaps/zoomin.xpm"
    #include "bitmaps/zoomout.xpm"
    #include "bitmaps/info.xpm"

    // navigator
    #include "bitmaps/north.xpm"
    #include "bitmaps/south.xpm"
    #include "bitmaps/east.xpm"
    #include "bitmaps/west.xpm"
    #include "bitmaps/center.xpm"
#endif

CellSet::CellSet()
{
  m_minX = m_minY = m_maxX = m_maxY = 0;
  SetSize(0, CELLSET_DEFAULT_MAX, 0, CELLSET_DEFAULT_MAX);
}

CellSet::CellSet(wxInt32 minX, wxInt32 maxX, wxInt32 minY, wxInt32 maxY)
{
  m_minX = m_minY = m_maxX = m_maxY = 0;
  SetSize(minX, maxX, minY, maxY);
}

CellSet::~CellSet()
{
  clear();
}

size_t CellSet::GetDimX()
{
  return (size_t)(m_maxX - m_minX) + 1;
}

size_t CellSet::GetDimY()
{
  return (size_t)(m_maxY - m_minY) + 1;
}

void CellSet::SetSize(wxInt32 minX, wxInt32 maxX, wxInt32 minY, wxInt32 maxY)
{
  CellSet temp = *this;

  m_minX = minX;
  m_maxX = maxX;
  m_minY = minY;
  m_maxY = maxY;

  Clear();

  for (size_t i=0; i<temp.size(); i++) {
    wxInt32 tX = temp.GetXFromIndex(i);
    wxInt32 tY = temp.GetYFromIndex(i);
    if ((tX >= m_minX) && (tX <= m_maxX) && (tY >= m_minY) && (tY <= m_maxY)) {
      SetCellValue(tX, tY, temp.GetCellValue(tX, tY));
    }
  }
}

void CellSet::SetCellValue(wxInt32 x, wxInt32 y, bool bVal)
{
  at(GetIndex(x, y)).m_bOn = bVal;
}

bool CellSet::GetCellValue(wxInt32 x, wxInt32 y)
{
  size_t ind = GetIndex(x, y);
  if (ind >= size()) {
    return false;
  } else {
    return at(ind).m_bOn;
  }
}

size_t CellSet::GetIndex(wxInt32 x, wxInt32 y)
{
  return ((y - m_minY) * GetDimX()) + (x - m_minX);
}

wxInt32 CellSet::GetXFromIndex(size_t ind)
{
  return (ind % GetDimX()) + m_minX;
}

wxInt32 CellSet::GetYFromIndex(size_t ind)
{
  return ((ind - (ind % GetDimX())) / GetDimX()) + m_minY;
}

void CellSet::Clear()
{
  clear();
  for (size_t i=0; i<(GetDimX()*GetDimY()); i++) {
    push_back(Cell());
  }
}


// --------------------------------------------------------------------------
// constants
// --------------------------------------------------------------------------

// IDs for the controls and the menu commands. Exluding those already defined
// by wxWidgets, such as wxID_NEW.
enum
{
    ID_INFO,
    ID_PENCIL,
    ID_RECTANGLE,
    ID_COLOR,
    ID_CLIPBOARD,
};

// --------------------------------------------------------------------------
// event tables and other macros for wxWidgets
// --------------------------------------------------------------------------

// Event tables
BEGIN_EVENT_TABLE(FiberedFractalTilerFrame, wxFrame)
    EVT_MENU            (wxID_NEW,          FiberedFractalTilerFrame::OnMenu)
    EVT_MENU            (wxID_OPEN,         FiberedFractalTilerFrame::OnOpen)
    EVT_MENU            (wxID_SAVE,         FiberedFractalTilerFrame::OnSave)
    EVT_MENU            (wxID_SAVEAS,       FiberedFractalTilerFrame::OnSaveAs)
    EVT_MENU            (wxID_EXIT,         FiberedFractalTilerFrame::OnMenu)
    EVT_MENU            (wxID_ZOOM_IN,      FiberedFractalTilerFrame::OnZoom)
    EVT_MENU            (wxID_ZOOM_OUT,     FiberedFractalTilerFrame::OnZoom)
    EVT_MENU            (ID_SETGRIDSIZE,    FiberedFractalTilerFrame::OnGridSize)
    EVT_MENU            (ID_COLOR,          FiberedFractalTilerFrame::OnColor)
    EVT_MENU            (wxID_ABOUT,        FiberedFractalTilerFrame::OnMenu)
    EVT_MENU            (wxID_HELP,         FiberedFractalTilerFrame::OnMenu)
    EVT_MENU            (ID_PENCIL,         FiberedFractalTilerFrame::OnMenu)
    EVT_MENU            (ID_RECTANGLE,      FiberedFractalTilerFrame::OnMenu)
    EVT_MENU            (ID_CLIPBOARD,      FiberedFractalTilerFrame::OnMenu)
    EVT_TEXT            (ID_CTRL_SCALE,     FiberedFractalTilerFrame::OnScale)
    EVT_RADIOBOX        (ID_RADIO_MODE,     FiberedFractalTilerFrame::OnMode)
    EVT_CHOICE          (ID_CTRL_QUAD,      FiberedFractalTilerFrame::OnQuadrant)
    EVT_BUTTON          (ID_BUTT_FORWARD,   FiberedFractalTilerFrame::OnForward)
    EVT_BUTTON          (ID_BUTT_RESET,     FiberedFractalTilerFrame::OnReset)
    EVT_BUTTON          (ID_BUTT_GEN,       FiberedFractalTilerFrame::OnGenerate)
    EVT_BUTTON          (ID_TILESET_BROWSE, FiberedFractalTilerFrame::OnBrowse)
    EVT_BUTTON          (ID_SEED_BROWSE,    FiberedFractalTilerFrame::OnBrowse)
    EVT_CLOSE           (                   FiberedFractalTilerFrame::OnClose)
END_EVENT_TABLE()

IMPLEMENT_CLASS(FiberedFractalTilerCanvas, wxScrolledWindow)
BEGIN_EVENT_TABLE(FiberedFractalTilerCanvas, wxScrolledWindow)
    EVT_PAINT           (             FiberedFractalTilerCanvas::OnPaint)
    EVT_SIZE            (             FiberedFractalTilerCanvas::OnSize)
    EVT_MOTION          (             FiberedFractalTilerCanvas::OnMouse)
    EVT_LEFT_DOWN       (             FiberedFractalTilerCanvas::OnMouse)
    EVT_LEFT_UP         (             FiberedFractalTilerCanvas::OnMouse)
    EVT_LEFT_DCLICK     (             FiberedFractalTilerCanvas::OnMouse)
    EVT_RIGHT_DOWN      (             FiberedFractalTilerCanvas::OnMouse)
    EVT_RIGHT_UP        (             FiberedFractalTilerCanvas::OnMouse)
    EVT_LEAVE_WINDOW    (             FiberedFractalTilerCanvas::OnMouse)
    EVT_ERASE_BACKGROUND(             FiberedFractalTilerCanvas::OnEraseBackground)
    EVT_MENU            (wxID_CUT,    FiberedFractalTilerCanvas::OnCut)
    EVT_MENU            (wxID_COPY,   FiberedFractalTilerCanvas::OnCopy)
    EVT_MENU            (wxID_PASTE,  FiberedFractalTilerCanvas::OnPaste)
END_EVENT_TABLE()


// Create a new application object
IMPLEMENT_APP(FiberedFractalTilerApp)


// ==========================================================================
// implementation
// ==========================================================================

// some shortcuts
#define ADD_TOOL(id, bmp, tooltip, help) \
    toolBar->AddTool(id, bmp, wxNullBitmap, false, wxDefaultCoord, wxDefaultCoord, (wxObject *)NULL, tooltip, help)


// --------------------------------------------------------------------------
// FiberedFractalTilerApp
// --------------------------------------------------------------------------

// 'Main program' equivalent: the program execution "starts" here
bool FiberedFractalTilerApp::OnInit()
{
    // create the main application window
    FiberedFractalTilerFrame *frame = new FiberedFractalTilerFrame();

    // show it and tell the application that it's our main window
    frame->Show(true);
    SetTopWindow(frame);

    // just for Motif
#ifdef __WXMOTIF__
    frame->UpdateInfoText();
#endif

    // enter the main message loop and run the app
    return true;
}

//  Definition of a basic '?' image to be used for the help button
static const char * help_xpm[] = {
"12 11 2 1",
"  c None",
". c #000000",
"    ....    ",
"   ......   ",
"  ...  ...  ",
"  ...  ...  ",
"      ...   ",
"     ...    ",
"     ...    ",
"     ...    ",
"            ",
"     ...    ",
"     ...    "};

static const char * clipboard_xpm[] = {
"12 11 2 1",
"  c None",
". c #000000",
"  ........  ",
"  .      .  ",
"  .      .  ",
"  .      .  ",
"  .      .  ",
"  .      .  ",
"  .      .  ",
"  .      .  ",
"  .      .  ",
"  ........  ",
"            "};

// --------------------------------------------------------------------------
// FiberedFractalTilerFrame
// --------------------------------------------------------------------------

#ifndef FIBERER_ONLY
  #define TITLE _T("Shape Tiler")
#else
  #define TITLE _T("Fibered Fractal Tiler")
#endif  // FIBERER_ONLY

// frame constructor
FiberedFractalTilerFrame::FiberedFractalTilerFrame() :
  wxFrame( (wxFrame *) NULL, wxID_ANY, TITLE, wxDefaultPosition ),
  m_cellSet(0, CELLSET_DEFAULT_MAX, 0, CELLSET_DEFAULT_MAX),
  m_helpController(wxHF_DEFAULT_STYLE, this)
{
    m_ptcScale        = NULL;
    m_ptcTilesetFile  = NULL;
    m_ptcSeedFile     = NULL;
    m_ptcMessageBox   = NULL;
    m_pccQuadrant     = NULL;
    m_pbuttonGenerate = NULL;
    m_pStreamRedirect = NULL;

    m_mode            = MODE_FRACTAL;
    m_drawMode        = DRAWMODE_PENCIL;
    m_stage           = 1;
    m_quadrant        = QUAD1;
    m_colCellOn       = *wxBLACK;

    m_penCellOn.SetColour(m_colCellOn);
    m_brushCellOn.SetColour(m_colCellOn);

    wxFileSystem::AddHandler(new wxArchiveFSHandler);
#ifndef FIBERER_ONLY
    m_helpController.Initialize("FiberedFractalTilerHelp.zip");
#else
    m_helpController.Initialize("FiberedFractalTilerHelp.zip");
#endif  // FIBERER_ONLY

    // frame icon
    SetIcon(wxICON(mondrian));

    // menu bar
    wxMenu *menuFile = new wxMenu(wxMENU_TEAROFF);
    wxMenu *menuView = new wxMenu(wxMENU_TEAROFF);
    wxMenu *menuHelp = new wxMenu(wxMENU_TEAROFF);

    menuFile->Append(wxID_NEW, wxEmptyString, _("Start a new shape"));
    menuFile->AppendSeparator();
    menuFile->Append(wxID_OPEN, wxEmptyString, _("Open an existing shape defintion"));
    menuFile->AppendSeparator();
    menuFile->Append(wxID_SAVE, wxEmptyString, _("Save shape definition"));
    menuFile->Append(wxID_SAVEAS, wxEmptyString, _("Save shape definition as.."));

    menuFile->AppendSeparator();
    menuFile->Append(wxID_EXIT);

    menuView->Append(wxID_ZOOM_IN, wxEmptyString, _("Zoom in"));
    menuView->Append(wxID_ZOOM_OUT, wxEmptyString, _("Zoom out"));

    menuView->AppendSeparator();
    menuView->Append(ID_SETGRIDSIZE, "Set grid dimensions", _("Manually set grid dimensions"));
#ifndef FIBERER_ONLY
    menuView->AppendSeparator();
    menuView->Append(ID_COLOR, "Get color value...", "Get the numeric value for a color");
#endif  // FIBERER_ONLY

    menuHelp->Append(wxID_ABOUT, _("&About\tCtrl-A"), _("Show about dialog"));
    menuHelp->AppendSeparator();
    menuHelp->Append(wxID_HELP, "Help\tF1", "Show help contents");

    wxMenuBar *menuBar = new wxMenuBar();
    menuBar->Append(menuFile, _("&File"));
    menuBar->Append(menuView, _("&View"));
    menuBar->Append(menuHelp, _("&Help"));
    SetMenuBar(menuBar);

    // tool bar
    wxBitmap tbBitmaps[8];

    tbBitmaps[0] = wxBITMAP(reset);
    tbBitmaps[1] = wxBITMAP(open);
    tbBitmaps[2] = wxBITMAP(zoomin);
    tbBitmaps[3] = wxBITMAP(zoomout);
    tbBitmaps[4] = wxBITMAP(info);
    tbBitmaps[5] = wxBITMAP(pencil);
    tbBitmaps[6] = wxBITMAP(rectangle);
    tbBitmaps[7] = wxBitmap(help_xpm);
//    tbBitmaps[8] = wxBitmap(clipboard_xpm);

    wxToolBar *toolBar = CreateToolBar();
    toolBar->SetMargins(5, 5);
    toolBar->SetToolBitmapSize(wxSize(16, 16));

    ADD_TOOL(wxID_NEW, tbBitmaps[0], wxGetStockLabel(wxID_NEW, wxSTOCK_NOFLAGS), _("Start a new shape"));
    ADD_TOOL(wxID_OPEN, tbBitmaps[1], wxGetStockLabel(wxID_OPEN, wxSTOCK_NOFLAGS), _("Open an existing shape definition"));

    toolBar->AddSeparator();
    ADD_TOOL(wxID_ZOOM_IN, tbBitmaps[2], wxGetStockLabel(wxID_ZOOM_IN, wxSTOCK_NOFLAGS), _("Zoom in"));
    ADD_TOOL(wxID_ZOOM_OUT, tbBitmaps[3], wxGetStockLabel(wxID_ZOOM_OUT, wxSTOCK_NOFLAGS), _("Zoom out"));
    toolBar->AddSeparator();
    toolBar->AddTool(ID_PENCIL,    _("Pencil"), tbBitmaps[5], _("Draw with pencil"), wxITEM_RADIO);
    toolBar->AddTool(ID_RECTANGLE, _("Rectangle"), tbBitmaps[6], _("Draw/erase rectangular region"), wxITEM_RADIO);
//    toolBar->AddTool(ID_CLIPBOARD, _("Clipboard"), tbBitmaps[8], _("Clipboard mode"), wxITEM_RADIO);
    toolBar->AddSeparator();
    wxString str1 = _T("About ");
    str1 += TITLE;
    wxString str2 = _T("Display ");
    str2 += TITLE;
    str2 += _T(" information");
    ADD_TOOL(wxID_ABOUT, tbBitmaps[4], str1, str2);
    toolBar->AddTool(wxID_HELP, _("Help"), tbBitmaps[7], _("Display help documentation"));
    toolBar->AddSeparator();

    toolBar->Realize();

    // status bar
    CreateStatusBar(1);
    SetStatusText(_(""));

    // We use two different panels to reduce flicker in wxGTK, because
    // some widgets (like wxStaticText) don't have their own X11 window,
    // and thus updating the text would result in a refresh of the canvas
    // if they belong to the same parent.

    wxSplitterWindow* pSplitWin = new wxSplitterWindow(this, wxID_ANY);

    wxPanel *panel1 = new wxPanel(pSplitWin, wxID_ANY);

    wxScrolledWindow *panel2 = new wxScrolledWindow(pSplitWin, wxID_ANY);
    panel2->SetScrollbars(20, 20, 50, 50);

    // canvas
    m_canvas = new FiberedFractalTilerCanvas(panel1, this, &m_cellSet);

    // info panel
    m_scale = 1;

#ifndef FIBERER_ONLY
    m_ptcScale        = new wxTextCtrl(panel2, ID_CTRL_SCALE, "1", wxDefaultPosition, wxSize(40, 17), wxTE_RIGHT);
#endif  // #ifndef FIBERER_ONLY
    m_ptcTilesetFile  = new wxTextCtrl(panel2, ID_CTRL_TSET, "", wxDefaultPosition, wxSize(350, 17));
    m_ptcSeedFile     = new wxTextCtrl(panel2, ID_CTRL_SEED, "", wxDefaultPosition, wxSize(350, 17));
    m_ptcMessageBox   = new wxTextCtrl(panel2, ID_CTRL_MSGS, "", wxDefaultPosition, wxSize(600, 80), wxTE_MULTILINE | wxTE_READONLY);

    wxArrayString arrChoices;
    arrChoices.Add("1");
    arrChoices.Add("2");
    arrChoices.Add("3");
    arrChoices.Add("4");
    m_pccQuadrant     = new wxChoice(panel2, ID_CTRL_QUAD, wxDefaultPosition, wxSize(40, 17), arrChoices);

#ifndef FIBERER_ONLY
    wxString arrCh[]  = {_("Shape Definition"), _("Self-Similar Fractal")};
    m_pradioMode      = new wxRadioBox(panel2, ID_RADIO_MODE, "Tileset Generation Mode", wxDefaultPosition, wxDefaultSize, 2, arrCh);
    m_pradioMode->SetSelection(m_mode);
#endif  // FIBERER_ONLY

    m_pbuttonGenerate = new wxButton(panel2, ID_BUTT_GEN, "Generate Tileset");

    UpdateInfoText();

    // component layout
    wxBoxSizer *sizer1            = new wxBoxSizer(wxVERTICAL);

    wxBoxSizer *sizer3            = new wxBoxSizer(wxVERTICAL);

    sizer1->Add( new wxStaticLine(panel1, wxID_ANY), 0, wxGROW );
    sizer1->Add( m_canvas, 1, wxGROW);
    sizer1->Add( new wxStaticLine(panel1, wxID_ANY), 0, wxGROW );

    panel1->SetSizer( sizer1 );
    sizer1->Fit( panel1 );

    wxBoxSizer *bottomSizer = new wxBoxSizer(wxVERTICAL);

    wxBoxSizer *topOfBottomSizer = new wxBoxSizer(wxHORIZONTAL);
#ifndef FIBERER_ONLY
    topOfBottomSizer->Add(m_pradioMode, 0, wxTOP | wxBOTTOM | wxLEFT, 5);
#endif  // FIBERER_ONLY

    wxStaticBox *staticBoxControls = new wxStaticBox(panel2, wxID_ANY, "Fractal Growth Simulation");
    wxStaticBoxSizer* staticBoxControlsSizer = new wxStaticBoxSizer(staticBoxControls, wxHORIZONTAL);
    staticBoxControlsSizer->Add(new wxStaticText(panel2, wxID_ANY, _("  Stage:")), 0, wxTOP, 5);
    m_ptcStage = new wxTextCtrl(panel2, ID_STAGE, wxString::Format("%d", m_stage), wxDefaultPosition, wxSize(20, 17), wxTE_READONLY | wxTE_RIGHT);
    staticBoxControlsSizer->Add(m_ptcStage, 0, wxTOP, 4);
 
    m_pbuttonForward = new wxButton(panel2, ID_BUTT_FORWARD, "Forward", wxDefaultPosition, wxSize(50, 19));
    staticBoxControlsSizer->AddSpacer(12);
    staticBoxControlsSizer->Add(m_pbuttonForward, 0, wxTOP, 3);

    m_pbuttonReset = new wxButton(panel2, ID_BUTT_RESET, "Reset", wxDefaultPosition, wxSize(50, 19));
    staticBoxControlsSizer->AddSpacer(8);
    staticBoxControlsSizer->Add(m_pbuttonReset, 0, wxTOP, 3);

#ifndef FIBERER_ONLY
    topOfBottomSizer->AddSpacer(3);
#endif  // FIBERER_ONLY
    topOfBottomSizer->Add(staticBoxControlsSizer, 0, wxTOP | wxBOTTOM | wxLEFT, 5);

    bottomSizer->Add(topOfBottomSizer);


#ifndef FIBERER_ONLY
    wxFlexGridSizer *flexSizer = new wxFlexGridSizer(4, 3, 7, 7);
    flexSizer->Add(new wxStaticText(panel2, wxID_ANY, _("  Scale:")));
    flexSizer->Add(m_ptcScale);
    flexSizer->AddSpacer(0);
#else
    wxFlexGridSizer *flexSizer = new wxFlexGridSizer(3, 3, 7, 7);
#endif  // FIBERER_ONLY

    flexSizer->Add(new wxStaticText(panel2, wxID_ANY, _("  Quadrant:")));
    flexSizer->Add(m_pccQuadrant);
    flexSizer->AddSpacer(0);

    flexSizer->Add(new wxStaticText(panel2, wxID_ANY, "  Tileset output file:"));
    flexSizer->Add(m_ptcTilesetFile);
    flexSizer->Add(new wxButton(panel2, ID_TILESET_BROWSE, "Browse...", wxDefaultPosition, wxSize(60, 19)));

    flexSizer->Add(new wxStaticText(panel2, wxID_ANY, "  Seed output file:"));
    flexSizer->Add(m_ptcSeedFile);
    flexSizer->Add(new wxButton(panel2, ID_SEED_BROWSE, "Browse...", wxDefaultPosition, wxSize(60, 19)));
    bottomSizer->Add(flexSizer);

    bottomSizer->Add(m_pbuttonGenerate, 0, wxTOP | wxBOTTOM | wxLEFT, 5);

    wxStaticBox* statBox = new wxStaticBox(panel2, wxID_ANY, "Messages:");
    wxStaticBoxSizer* msgSizer = new wxStaticBoxSizer(statBox,wxHORIZONTAL);
    msgSizer->Add(m_ptcMessageBox);

    bottomSizer->Add(msgSizer, 0, wxTOP | wxBOTTOM | wxLEFT, 7);

    panel2->SetSizer(bottomSizer);
    bottomSizer->Fit(panel2);

    pSplitWin->SetSashGravity(0.5);
    pSplitWin->SplitHorizontally(panel1, panel2);

    sizer3->Add(pSplitWin, 1, wxGROW);
    SetSizer(sizer3);

    sizer3->Fit(this);

    m_pStreamRedirect = new wxStreamToTextRedirector(m_ptcMessageBox);

    wxCommandEvent evt;
    OnMode(evt);

    SetSize(800, 800);
    pSplitWin->SetSashPosition(400);
}

FiberedFractalTilerFrame::~FiberedFractalTilerFrame()
{
  delete m_pStreamRedirect;
}

int FiberedFractalTilerFrame::GetQuadrant()
{
  if (m_mode == MODE_SHAPE) {
    return QUAD1;
  } else {
    return m_quadrant;
  }
}

void FiberedFractalTilerFrame::UpdateInfoText()
{
#ifndef FIBERER_ONLY
  m_ptcScale->SetValue(wxString::Format("%d", m_scale));
#endif  // FIBERER_ONLY
  m_ptcTilesetFile->SetValue(m_fnTilesetFile.GetFullPath());
  m_ptcSeedFile->SetValue(m_fnSeedFile.GetFullPath());
}

// Enable or disable tools and menu entries according to the current
// state. See also wxEVT_UPDATE_UI events for a slightly different
// way to do this.
void FiberedFractalTilerFrame::UpdateUI()
{
    // zooming
    int cellsize = m_canvas->GetCellSize();
    GetToolBar()->EnableTool(wxID_ZOOM_IN,  cellsize < 32);
    GetToolBar()->EnableTool(wxID_ZOOM_OUT, cellsize > 1);
    GetMenuBar()->Enable(wxID_ZOOM_IN,  cellsize < 32);
    GetMenuBar()->Enable(wxID_ZOOM_OUT, cellsize > 1);
}

// Event handlers -----------------------------------------------------------

// OnMenu handles all events which don't have their own event handler
void FiberedFractalTilerFrame::OnMenu(wxCommandEvent& event)
{
    switch (event.GetId())
    {
        case wxID_NEW:
        {
            m_cellSet.Clear();
            m_fnShapeDefFile.Clear();
            m_fnTilesetFile.Clear();
            m_fnSeedFile.Clear();
            m_canvas->Scroll(0,10000);
            UpdateInfoText();
            Refresh();
            break;
        }
        case wxID_ABOUT:
        {
#ifndef FIBERER_ONLY
            wxMessageDialog abt(this, "Shape Tiler is a program that generates TAM tilesets which assemble user defined\nshapes or for fibered versions of a subset of discrete self-similar fractals.  The\ntileset formats are compatible with the 3DTrend simulator.\n\nIt was written by Matt Patitz (mpatitz@yahoo.com) and is based on work\nby Matt Patitz, Scott Summers.\n\n",
                                "About Shape Tiler", wxOK);
#else
            wxMessageDialog abt(this, "Fibered Fractal Tiler is a program that generates TAM tilesets which assemble\nfibered versions of user defined 'nice' discrete self-similar fractals.  The\ntileset formats are compatible with the 3DTrend simulator.  Please see the\nhelp documentation for more information.\n\nIt was written by Matt Patitz (mpatitz@yahoo.com) and is based on work\nby Matt Patitz, Scott Summers.\n\n",
                                "About Fibered Fractal Tiler", wxOK);
#endif  // FIBERER_ONLY
            abt.ShowModal();
            break;
        }
        case wxID_HELP:
        {
          m_helpController.DisplayContents();
          break;
        }
        case wxID_EXIT:
        {
            // true is to force the frame to close
            Close(true);
            break;
        }
        case ID_PENCIL:
        {
          m_canvas->SetDrawingMode(DRAWMODE_PENCIL);
          break;
        }
        case ID_RECTANGLE:
        {
          m_canvas->SetDrawingMode(DRAWMODE_RECTANGLE);
          break;
        }
        case ID_CLIPBOARD:
        {
          m_canvas->SetDrawingMode(DRAWMODE_CLIPBOARD);
          break;
        }
    }
}

void FiberedFractalTilerFrame::OnOpen(wxCommandEvent& WXUNUSED(event))
{
    wxFileDialog filedlg(this,
                         _("Choose a file to open"),
                         wxEmptyString,
                         wxEmptyString,
                         _("Point sets (*.pts)|*.pts|All files (*.*)|*.*"),
                         wxFD_OPEN | wxFD_FILE_MUST_EXIST);
    if (filedlg.ShowModal() == wxID_OK)
    {
      if (filedlg.GetPath().length() > 0) {
        m_ptcMessageBox->Clear();
        m_pointSet.Clear();

        std::string sPath = filedlg.GetPath().c_str();
        if (m_pointSet.GetPointsFromFile(sPath) && (m_pointSet.size() > 0)) {

          m_cellSet.Clear();

          wxInt32 minX, maxX, minY, maxY;
          minX = maxX = m_pointSet[0].GetX();
          minY = maxY = m_pointSet[0].GetY();
          for (size_t i=0; i<m_pointSet.size(); i++) {
            if (m_pointSet[i].GetX() < minX) minX = m_pointSet[i].GetX();
            if (m_pointSet[i].GetX() > maxX) maxX = m_pointSet[i].GetX();
            if (m_pointSet[i].GetY() < minY) minY = m_pointSet[i].GetY();
            if (m_pointSet[i].GetY() > maxY) maxY = m_pointSet[i].GetY();
          }

          //  Assume this is a fractal
          m_mode = MODE_FRACTAL;
          if (minX < 0) {
            if (minY < 0) {
              m_quadrant = QUAD3;
            } else {
              m_quadrant = QUAD4;
            }
            m_pccQuadrant->SetSelection(m_quadrant);
          } else if (minY < 0) {
            m_quadrant = QUAD2;
            m_pccQuadrant->SetSelection(m_quadrant);
          }
#ifndef FIBERER_ONLY
          m_pradioMode->SetSelection(m_mode);
#endif  // FIBERER_ONLY
          wxCommandEvent evt;
          OnMode(evt);

          if ((m_cellSet.m_minX > minX) || (m_cellSet.m_maxX < maxX) || (m_cellSet.m_minY > minY) || (m_cellSet.m_maxY < maxY)) {
            m_cellSet.SetSize(minX-20, maxX+20, minY-20, maxY+20);
            AddMessage(wxString::Format("Automatically growing grid size to (%d, %d) - (%d, %d) to fit point set\n", m_cellSet.m_minX, m_cellSet.m_minY, m_cellSet.m_maxX, m_cellSet.m_maxY));
            m_canvas->UpdateCanvasSize();
          }

          for (size_t i=0; i<m_pointSet.size(); i++) {
            m_cellSet.SetCellValue(m_pointSet[i].GetX(), m_pointSet[i].GetY(), true);
          }

          m_fnShapeDefFile  = sPath;
          m_fnTilesetFile   = m_fnShapeDefFile;
          m_fnTilesetFile.SetExt("tds");
          m_fnSeedFile      = m_fnShapeDefFile;
          m_fnSeedFile.SetExt("tdp");

          UpdateInfoText();
          m_canvas->Refresh();

        } else {
          wxString sErrMsg = wxString::Format("Failed to read point set from input file '%s'", filedlg.GetPath());
          wxMessageBox(sErrMsg, "Error reading input file", wxICON_ERROR, this);
        }
      }
    }
}

void FiberedFractalTilerFrame::OnSave(wxCommandEvent& WXUNUSED(event))
{
  if (m_fnShapeDefFile.GetFullPath().length() == 0) {
    wxMessageBox("Please select 'Save As..' to specify a file name.", "No file name specified", wxICON_ERROR);
    return;
  }

  m_pointSet.Clear();

  for (size_t i=0; i<m_cellSet.size(); i++) {
    if (m_cellSet[i].m_bOn) {
      Point pt(m_cellSet.GetXFromIndex(i), m_cellSet.GetYFromIndex(i));
      m_pointSet.push_back(pt);
    }
  }
  
  if (!m_pointSet.PrintToFile(m_fnShapeDefFile.GetFullPath().c_str())) {
    wxMessageBox("Failed to write point set to file", "Error saving point set", wxICON_ERROR);
  }
}

void FiberedFractalTilerFrame::OnSaveAs(wxCommandEvent& event)
{
  wxFileDialog filedlg(this,
                       _("Enter an output file name"),
                       wxEmptyString,
                       wxEmptyString,
                       _("Point sets (*.pts)|*.pts|All files (*.*)|*.*"),
                       wxFD_SAVE);

  if (filedlg.ShowModal() == wxID_OK)
  {
    if (filedlg.GetPath().length() > 0) {
      m_fnShapeDefFile = filedlg.GetPath();
      OnSave(event);

      if (m_fnTilesetFile.GetFullPath().length() <= 0) {
        m_fnTilesetFile = m_fnShapeDefFile;
        m_fnTilesetFile.SetExt("tds");
      }

      if (m_fnSeedFile.GetFullPath().length() <= 0) {
        m_fnSeedFile = m_fnShapeDefFile;
        m_fnSeedFile.SetExt("tdp");
      }

      UpdateInfoText();
    }
  }
}

void FiberedFractalTilerFrame::OnZoom(wxCommandEvent& event)
{
    int cellsize = m_canvas->GetCellSize();

    if ((event.GetId() == wxID_ZOOM_IN) && cellsize < 32)
    {
        m_canvas->SetCellSize(cellsize * 2);
        UpdateUI();
    }
    else if ((event.GetId() == wxID_ZOOM_OUT) && cellsize > 1)
    {
        m_canvas->SetCellSize(cellsize / 2);
        UpdateUI();
    }
}

void FiberedFractalTilerFrame::OnGridSize(wxCommandEvent& WXUNUSED(event))
{
  wxTextEntryDialog ted(this, "Enter a positive integer for the\nnumber of rows and columns:", "Set grid dimensions", wxString::Format("%d", m_cellSet.GetDimX()));
  if (ted.ShowModal() == wxID_OK) {
    wxString sVal = ted.GetValue();
    long lVal;
    if (sVal.ToLong(&lVal) && (lVal > 0)) {
      m_cellSet.SetSize(0, (int)lVal, 0, (int)lVal);
      m_canvas->UpdateCanvasSize();
    } else {
      wxMessageBox("Error: Invalid value entered and ignored.", "Invalid entry", wxICON_ERROR);
    }
  }
}

void FiberedFractalTilerFrame::OnColor(wxCommandEvent& WXUNUSED(event))
{
  wxColourDialog cd(this);
  if (cd.ShowModal() == wxID_OK) {
    m_colCellOn = cd.GetColourData().GetColour();
    m_penCellOn.SetColour(m_colCellOn);
    m_brushCellOn.SetColour(m_colCellOn);
    wxMessageBox(wxString::Format("Color value: %s", m_colCellOn.GetAsString(wxC2S_CSS_SYNTAX)), "Color Value");
    Refresh();
  }
}

void FiberedFractalTilerFrame::OnScale(wxCommandEvent& WXUNUSED(event))
{
  if (m_ptcScale) {
    wxString sVal = m_ptcScale->GetValue();
    long lVal = 1;
    if (!sVal.ToLong(&lVal) || (lVal < 1)) {
      wxMessageBox("Value must be a number greater than or equal to 1", "Invalid input", wxICON_ERROR, this);
      m_ptcScale->SetValue(wxString::Format("%d", m_scale));
    } else {
      m_scale = lVal;
    }
  }
}

void FiberedFractalTilerFrame::OnMode(wxCommandEvent& WXUNUSED(event))
{
#ifndef FIBERER_ONLY
  m_mode = m_pradioMode->GetSelection();
#else
  m_mode = MODE_FRACTAL;
#endif  // FIBERER_ONLY
  if (m_mode == MODE_SHAPE) {
    m_cellSet.SetSize(0, CELLSET_DEFAULT_MAX, 0, CELLSET_DEFAULT_MAX);
    m_pbuttonForward->Enable(false);
    m_pbuttonReset->Enable(false);
    m_ptcStage->Enable(false);
    m_ptcScale->Enable(true);
    m_pccQuadrant->Enable(false);
    m_canvas->Scroll(0, 9999999);
    Refresh();
  } else if (m_mode == MODE_FRACTAL) {
    m_pbuttonForward->Enable(true);
    m_pbuttonReset->Enable(true);
    m_ptcStage->Enable(true);
#ifndef FIBERER_ONLY
    m_ptcScale->Enable(false);
#endif  // FIBERER_ONLY
    m_pccQuadrant->Enable(true);
    m_pccQuadrant->SetSelection(m_quadrant);
    wxCommandEvent evt;
    OnQuadrant(evt);
  }
}

void FiberedFractalTilerFrame::OnQuadrant(wxCommandEvent& WXUNUSED(event))
{
  m_quadrant = m_pccQuadrant->GetSelection();
  m_stage = 1;
  m_ptcStage->SetValue(wxString::Format("%d", m_stage));
  if (m_quadrant == QUAD1) {
    m_cellSet.SetSize(0, CELLSET_DEFAULT_MAX, 0, CELLSET_DEFAULT_MAX);
    m_canvas->Scroll(0, 9999999);
  } else if (m_quadrant == QUAD2) {
    m_cellSet.SetSize(0, CELLSET_DEFAULT_MAX, -CELLSET_DEFAULT_MAX, 0);
    m_canvas->Scroll(0, 0);
  } else if (m_quadrant == QUAD3) {
    m_cellSet.SetSize(-CELLSET_DEFAULT_MAX, 0, -CELLSET_DEFAULT_MAX, 0);
    m_canvas->Scroll(9999999, 0);
  } else if (m_quadrant == QUAD4) {
    m_cellSet.SetSize(-CELLSET_DEFAULT_MAX, 0, 0, CELLSET_DEFAULT_MAX);
    m_canvas->Scroll(9999999, 9999999);
  }
  Refresh();
}

void FiberedFractalTilerFrame::OnTilesetFile(wxCommandEvent& WXUNUSED(event))
{
  if (m_ptcTilesetFile) {

  }
}

void FiberedFractalTilerFrame::OnSeedFile(wxCommandEvent& WXUNUSED(event))
{
  if (m_ptcSeedFile) {

  }
}

void FiberedFractalTilerFrame::OnForward(wxCommandEvent& WXUNUSED(event))
{
  if (!m_cellSet.GetCellValue(0, 0)) {
    wxMessageBox("This is not a valid self-similar fractal:\nThe origin point (0, 0) must be selected.", "Invalid Fractal", wxICON_ERROR);
    return;
  }

  if (m_stage == 1) {   //  Save cellSet for stage 1
    m_cellSetStage1 = m_cellSet;
  }

  //  Use offsets of all stage 1 points from origin to make copies of existing cells
  double xMax = 0;
  double yMax = 0;
  for (size_t i=0; i<m_cellSetStage1.size(); i++) {
    if (m_cellSetStage1.at(i).m_bOn) {
      if (m_quadrant == QUAD1) {
        if (m_cellSetStage1.GetXFromIndex(i) > xMax) xMax = m_cellSetStage1.GetXFromIndex(i);
        if (m_cellSetStage1.GetYFromIndex(i) > yMax) yMax = m_cellSetStage1.GetYFromIndex(i);
      } else if (m_quadrant == QUAD2) {
        if (m_cellSetStage1.GetXFromIndex(i) > xMax) xMax = m_cellSetStage1.GetXFromIndex(i);
        if (m_cellSetStage1.GetYFromIndex(i) < yMax) yMax = m_cellSetStage1.GetYFromIndex(i);
      } else if (m_quadrant == QUAD3) {
        if (m_cellSetStage1.GetXFromIndex(i) < xMax) xMax = m_cellSetStage1.GetXFromIndex(i);
        if (m_cellSetStage1.GetYFromIndex(i) < yMax) yMax = m_cellSetStage1.GetYFromIndex(i);
      } else {
        if (m_cellSetStage1.GetXFromIndex(i) < xMax) xMax = m_cellSetStage1.GetXFromIndex(i);
        if (m_cellSetStage1.GetYFromIndex(i) > yMax) yMax = m_cellSetStage1.GetYFromIndex(i);
      }
    }
  }

  if (m_stage == 1) {
    m_maxXOn = xMax;
    m_maxYOn = yMax;
  }

  //  Increment stage
  m_stage++;

  double xTrans = pow(xMax+1, m_stage-1);
  double yTrans = pow(yMax+1, m_stage-1);

  size_t xDim = (m_maxXOn + (xMax * xTrans));
  size_t yDim = (m_maxYOn + (yMax * yTrans));

  if (m_quadrant == QUAD2) {
    yTrans = -abs(pow(yMax-1, m_stage-1));
    yDim = (m_maxYOn - (yMax * yTrans));
  } else if (m_quadrant == QUAD3) {
    xTrans = -abs(pow(xMax-1, m_stage-1));
    xDim = (m_maxXOn - (xMax * xTrans));
    yTrans = -abs(pow(yMax-1, m_stage-1));
    yDim = (m_maxYOn - (yMax * yTrans));
  } else if (m_quadrant == QUAD4) {
    xTrans = -abs(pow(xMax-1, m_stage-1));
    xDim = (m_maxXOn - (xMax * xTrans));
  }

  if ((abs((long)xDim) > m_cellSet.GetDimX()) || (abs((long)yDim) > m_cellSet.GetDimY())) {
    if (m_quadrant == QUAD1) {
      m_cellSet.SetSize(0, xDim + 50, 0, yDim + 50);
    } else if (m_quadrant == QUAD2) {
      m_cellSet.SetSize(0, xDim + 50, (0 - yDim) - 50, 0);
    } else if (m_quadrant == QUAD3) {
      m_cellSet.SetSize((0 - xDim) - 50, 0, (0 - yDim) - 50, 0);
    } else {
      m_cellSet.SetSize((0 - xDim) - 50, 0, 0, yDim + 50);
    }
    m_canvas->UpdateCanvasSize();
  }

  CellSet csTemp = m_cellSet;
  for (size_t j=0; j<m_cellSetStage1.size(); j++) {
    if (m_cellSetStage1.at(j).m_bOn) {
      for (size_t i=0; i<csTemp.size(); i++) {
        if (csTemp.at(i).m_bOn) {
          if (m_quadrant == QUAD1) {
            wxInt32 x = csTemp.GetXFromIndex(i) + (m_cellSetStage1.GetXFromIndex(j) * xTrans);
            wxInt32 y = csTemp.GetYFromIndex(i) + (m_cellSetStage1.GetYFromIndex(j) * yTrans);
            m_cellSet.SetCellValue(x, y, true);
            if (x > m_maxXOn) m_maxXOn = x;
            if (y > m_maxYOn) m_maxYOn = y;
          } else if (m_quadrant == QUAD2) {
            wxInt32 x = csTemp.GetXFromIndex(i) + (m_cellSetStage1.GetXFromIndex(j) * xTrans);
            wxInt32 y = csTemp.GetYFromIndex(i) - (m_cellSetStage1.GetYFromIndex(j) * yTrans);
            m_cellSet.SetCellValue(x, y, true);
            if (x > m_maxXOn) m_maxXOn = x;
            if (y < m_maxYOn) m_maxYOn = y;
          } else if (m_quadrant == QUAD3) {
            wxInt32 x = csTemp.GetXFromIndex(i) - (m_cellSetStage1.GetXFromIndex(j) * xTrans);
            wxInt32 y = csTemp.GetYFromIndex(i) - (m_cellSetStage1.GetYFromIndex(j) * yTrans);
            m_cellSet.SetCellValue(x, y, true);
            if (x < m_maxXOn) m_maxXOn = x;
            if (y < m_maxYOn) m_maxYOn = y;
          } else {
            wxInt32 x = csTemp.GetXFromIndex(i) - (m_cellSetStage1.GetXFromIndex(j) * xTrans);
            wxInt32 y = csTemp.GetYFromIndex(i) + (m_cellSetStage1.GetYFromIndex(j) * yTrans);
            m_cellSet.SetCellValue(x, y, true);
            if (x < m_maxXOn) m_maxXOn = x;
            if (y > m_maxYOn) m_maxYOn = y;
          }
        }
      }
    }
  }

  m_ptcStage->SetValue(wxString::Format("%d", m_stage));
  Refresh();
}

void FiberedFractalTilerFrame::OnReset(wxCommandEvent& WXUNUSED(event))
{
  if (m_stage == 1) {
    m_cellSetStage1 = m_cellSet;
  } else {
    m_cellSet.clear();
    m_cellSet.SetSize(0, CELLSET_DEFAULT_MAX, 0, CELLSET_DEFAULT_MAX);
    m_cellSet = m_cellSetStage1;
    m_stage   = 1;

    m_canvas->UpdateCanvasSize();

    m_ptcStage->SetValue(wxString::Format("%d", m_stage));
    Refresh();
  }
}

void FiberedFractalTilerFrame::OnBrowse(wxCommandEvent& event)
{
  if (event.GetId() == ID_TILESET_BROWSE) {
    wxFileDialog fdlg(this, "Tileset file", wxEmptyString, wxEmptyString, "TDS files (*.tds)|*.tds");
    if (fdlg.ShowModal() == wxID_OK) {
      m_fnTilesetFile = fdlg.GetPath();
      m_ptcTilesetFile->SetValue(m_fnTilesetFile.GetFullPath());
    }
  } else {
    wxFileDialog fdlg(this, "Seed file", wxEmptyString, wxEmptyString, "TDP files (*.tdp)|*.tdp");
    if (fdlg.ShowModal() == wxID_OK) {
      m_fnSeedFile = fdlg.GetPath();
      m_ptcSeedFile->SetValue(m_fnSeedFile.GetFullPath());
    }
  }
}

void FiberedFractalTilerFrame::AddMessage(wxString sMsg, bool bClear)
{
  if (bClear) {
    m_ptcMessageBox->Clear();
    m_ptcMessageBox->Refresh();
  }
  m_ptcMessageBox->AppendText(sMsg);
}

void FiberedFractalTilerFrame::OnGenerate(wxCommandEvent& WXUNUSED(event))
{
  if (!m_pbuttonGenerate) {
    return;
  }

  if ((m_mode == MODE_FRACTAL) && (m_stage > 1)) {
    wxString msg;
    msg += "The fractal is not in stage 1.  To generate a tileset ";
    msg += "from the first stage, press 'Cancel' and then the 'Reset' ";
    msg += "button before generating.  Otherwise, to continue using the ";
    msg += "current stage, press 'OK'.";
    wxMessageDialog dlg(this, msg, "Confirm tileset generation", wxOK | wxCANCEL);
    if (dlg.ShowModal() != wxID_OK) {
      return;
    }
  }

  AddMessage(_("Generating tileset...\n"), true);

  wxString desc;
  m_pointSet.Clear();

  for (size_t i=0; i<m_cellSet.size(); i++) {
    if (m_cellSet[i].m_bOn) {
      Point pt(m_cellSet.GetXFromIndex(i), m_cellSet.GetYFromIndex(i));
      m_pointSet.push_back(pt);
      //wxString sCoord = wxString::Format("(%d,%d) ", m_cellSet.GetXFromIndex(i), m_cellSet.GetYFromIndex(i));
      //desc += sCoord;
    }
  }

  if (m_pointSet.size() <= 0) {
    AddMessage("No points have been selected.  Please select a set of connected points.\n");
    AddMessage(_("Tileset generation failed\n"));
    return;
  } else {
    AddMessage(wxString::Format("%d points selected\n", m_pointSet.size()));
  }

  wxString sTilesetFile = m_ptcTilesetFile->GetValue();
  if (sTilesetFile.length() == 0) {
    AddMessage("An output file for the tileset must be specified.\n");
    AddMessage(_("Tileset generation failed\n"));
    return;
  }

  TileSet ts;

  if (m_mode == MODE_SHAPE) {
    //  Arbitrarily set the first point in the set as the origin
    m_pointSet.SetOrigin(0);
    AddMessage(wxString::Format("Point (%d, %d) used as origin\n", m_pointSet[0].GetX(), m_pointSet[0].GetY()));

    //  Verify connectedness
    if (!m_pointSet.CalculateInputs()) {
      AddMessage("The points are not completely connected, so a tileset cannot be generated.\n");
      AddMessage(_("Tileset generation failed\n"));
      return;
    } else {
      AddMessage("Points are connected!\n");
    }

    m_pointSet.CalculateOutputs();

    if (!GenerateTilesForShape(&m_pointSet, &ts, m_scale)) {
      AddMessage("There has been an error generating a tileset to construct the specified shape.\n");
      AddMessage(_("Tileset generation failed\n"));
      return;
    }
  } else if (m_mode == MODE_FRACTAL) {
    Point zz(0,0);
    int loc = INVALID_CELL;
    for (size_t i=0; i<m_pointSet.size(); i++) {
      if (m_pointSet[i] == zz) {
        loc = i;
      }
      m_pointSet[i].SetNum(i);
    }

    if (loc == INVALID_CELL) {
      AddMessage("The origin point (0, 0) must be selected to create a valid fractal definition.\n");
      AddMessage(_("Tileset generation failed\n"));
      return;
    } else {
      AddMessage(wxString::Format("Point (%d, %d) used as origin\n", m_pointSet[loc].GetX(), m_pointSet[loc].GetY()));
      m_pointSet.SetOrigin(loc);
    }

    if (!GenerateFiberedSSFTileset(&m_pointSet, &ts, m_quadrant)) {
      AddMessage(_("Tileset generation failed\n"));
      return;
    }
  } else {
    AddMessage("Unknown tile generation mode.\n");
    AddMessage(_("Tileset generation failed\n"));
    return;
  }

  AddMessage(wxString::Format("Tileset created with %d tiletypes\n", ts.size()));

  if (ts.size() > 7000) {
    wxString sWarn = wxString::Format(_T("Warning:  The tileset you have created consists of %d unique tiletypes.\n"), ts.size());
    sWarn         +=                  _T("That is a very large number and may be very difficult to load in the simulator.");
    wxMessageBox(sWarn, _T("Large Tileset Warning"), wxICON_WARNING);
  }

  if (!ts.PrintToFile(sTilesetFile.c_str(), TREND_TILETYPES)) {
    AddMessage("Failed to write tiletype definitions to output file.\n");
    AddMessage(_("Tileset generation failed\n"));
    return;
  } else {
    m_fnTilesetFile = sTilesetFile;
    wxString sMsg = wxString::Format("Tileset successfully written to output file '%s'\n", sTilesetFile);
    AddMessage(sMsg);
  }

  bool bWroteSeed = false;
  wxString sSeedFile = m_ptcSeedFile->GetValue();
  if (sSeedFile.length() > 0) {
    std::ofstream out(sSeedFile.c_str());
    if (out.is_open()) {
      if (m_mode == MODE_SHAPE) {
        wxString seedLine = wxString::Format("SEED %d %d\n", m_pointSet[m_pointSet.GetOrigin()].GetX()*m_scale, m_pointSet[m_pointSet.GetOrigin()].GetY()*m_scale);
        out << seedLine.c_str();
      } else if (m_mode == MODE_FRACTAL) {
        wxString seedLine = wxString::Format("SEED 7 7\n");
        out << seedLine.c_str();
      }
      out.close();
      bWroteSeed = true;
    }
  }
  if (!bWroteSeed) {
    AddMessage("Failed to write seed definition file.\n");
    AddMessage(_("Seed file generation failed\n"));
  } else {
    m_fnSeedFile = sSeedFile;
    wxString sMsg = wxString::Format("Seed successfully written to seed file '%s'\n", sSeedFile);
    AddMessage(sMsg);
  }

  AddMessage(_("Tileset generation completed successfully."));
}

void FiberedFractalTilerFrame::OnClose(wxCloseEvent& WXUNUSED(event))
{
  Destroy();
}

bool FiberedFractalTilerFrame::GenerateTilesForShape(PointSet* pPS, TileSet* pTS, long scale)
{
  if (scale > 1) {
    //  Create the 'filler' tiletype that fills in the non-axes positions of each (scale x scale) square
    Tile t;
    for (int i=NORTH; i<=WEST; i++) {
      t.m_sides[i].m_strength   = 1;
      t.m_sides[i].m_label      = FILLER_LABEL;
    }
    pTS->push_back(t);
  }

  //  Create the counter tiles for the axes
  for (size_t i=0; i<pPS->size(); i++) {

    int dirInput = pPS->at(i).GetInputDir();

    wxString sOutput  = wxString::Format("[%d,%d]", pPS->at(i).GetX(), pPS->at(i).GetY());
    wxString sInput   = NODISPLAY;
    if (pPS->at(i).GetInput()) {
      sInput = wxString::Format("[%d,%d]%s", pPS->at(i).GetInput()->GetX(), pPS->at(i).GetInput()->GetY(), DirectionStringFromInt(OppositeDirection(pPS->at(i).GetInputDir())).c_str());
    }

    //  The (0,0) tile for this type
    Tile t0;
    if (pPS->GetOrigin() == (int)i) {
      t0.SetName("SEED");
    }

    t0.m_display                  = sOutput;

    t0.m_sides[NORTH].m_strength  = (scale>1) ? 2
                                              : (dirInput==NORTH) ? 2
                                                                  : (pPS->at(i).GetOutput(NORTH)) ? 2
                                                                                                  : 0;
    t0.m_sides[NORTH].m_label     = (scale>1) ? sOutput + "," + itos(0)
                                              : (dirInput==NORTH) ? sInput
                                                                  : (pPS->at(i).GetOutput(NORTH)) ? sOutput + DirectionStringFromInt(NORTH)
                                                                                                  : NODISPLAY;

    t0.m_sides[EAST].m_strength   = (scale>1) ? 2
                                              : (dirInput==EAST) ? 2
                                                                 : (pPS->at(i).GetOutput(EAST)) ? 2
                                                                                                : 0;
    t0.m_sides[EAST].m_label      = (scale>1) ? sOutput + "," + itos(0)
                                              : (dirInput==EAST) ? sInput
                                                                 : (pPS->at(i).GetOutput(EAST)) ? sOutput + DirectionStringFromInt(EAST)
                                                                                                : NODISPLAY;

    t0.m_sides[SOUTH].m_strength  = (dirInput==SOUTH) ? 2
                                                      : (pPS->at(i).GetOutput(SOUTH)) ? 2
                                                                                      : 0;
    t0.m_sides[SOUTH].m_label     = (dirInput==SOUTH) ? sInput
                                                      : (pPS->at(i).GetOutput(SOUTH)) ? sOutput + DirectionStringFromInt(SOUTH)
                                                                                      : NODISPLAY;

    t0.m_sides[WEST].m_strength   = (dirInput==WEST) ? 2
                                                     : (pPS->at(i).GetOutput(WEST)) ? 2
                                                                                    : 0;
    t0.m_sides[WEST].m_label      = (dirInput==WEST) ? sInput
                                                     : (pPS->at(i).GetOutput(WEST)) ? sOutput + DirectionStringFromInt(WEST)
                                                                                    : NODISPLAY;

    pTS->push_back(t0);

    for (long s=1; s<scale; s++) {

      Tile t;
      t.m_display                 = sOutput + "," + itos(s);
      
      //  Create the tiles for the y-axis
      t.m_sides[NORTH].m_strength = (s<scale-1) ? 2
                                                : (dirInput==NORTH) ? 2
                                                                    : (pPS->at(i).GetOutput(NORTH)) ? 2
                                                                                                    : 0;
      t.m_sides[NORTH].m_label    = (s<scale-1) ? sOutput + "," + itos(s)
                                                : (dirInput==NORTH) ? sInput 
                                                                    : (pPS->at(i).GetOutput(NORTH)) ? sOutput + DirectionStringFromInt(NORTH)
                                                                                                    : NODISPLAY;

      t.m_sides[EAST].m_strength  = 1;
      t.m_sides[EAST].m_label     = FILLER_LABEL;

      t.m_sides[SOUTH].m_strength = 2;
      t.m_sides[SOUTH].m_label    = sOutput + "," + itos(s-1);

      t.m_sides[WEST].m_strength  = 0;
      t.m_sides[WEST].m_label     = NODISPLAY;

      pTS->push_back(t);

      //  Create the tiles for the x-axis
      t.m_sides[EAST].m_strength = (s<scale-1) ? 2
                                                : (dirInput==EAST) ? 2
                                                                    : (pPS->at(i).GetOutput(EAST)) ? 2
                                                                                                    : 0;
      t.m_sides[EAST].m_label    = (s<scale-1) ? sOutput + "," + itos(s)
                                                : (dirInput==EAST) ? sInput 
                                                                    : (pPS->at(i).GetOutput(EAST)) ? sOutput + DirectionStringFromInt(EAST)
                                                                                                    : NODISPLAY;

      t.m_sides[NORTH].m_strength  = 1;
      t.m_sides[NORTH].m_label     = FILLER_LABEL;

      t.m_sides[WEST].m_strength = 2;
      t.m_sides[WEST].m_label    = sOutput + "," + itos(s-1);

      t.m_sides[SOUTH].m_strength  = 0;
      t.m_sides[SOUTH].m_label     = NODISPLAY;

      pTS->push_back(t);
    }
  }
  return true;
}


// --------------------------------------------------------------------------
// FiberedFractalTilerCanvas
// --------------------------------------------------------------------------

// canvas constructor
FiberedFractalTilerCanvas::FiberedFractalTilerCanvas(wxWindow *parent, FiberedFractalTilerFrame *pFrame, CellSet* pCellSet)
          : wxScrolledWindow(parent, ID_CANVAS_WIN, wxDefaultPosition, wxDefaultSize, wxFULL_REPAINT_ON_RESIZE)
{
    m_pFrame      = pFrame;
    m_cellsize    = 16;
    m_status      = MOUSE_NOACTION;
    m_pCellSet    = pCellSet;

    m_mi          = INVALID_CELL;
    m_mj          = INVALID_CELL;

    // reduce flicker if wxEVT_ERASE_BACKGROUND is not available
    SetBackgroundColour(*wxLIGHT_GREY);
//    SetBackgroundStyle(wxBG_STYLE_CUSTOM);

    // Create a Selection Box (Rubber Band)
	  m_RubberBand = wxRubberBand(this);
    m_rubberbandPen.SetStyle(wxSOLID);
    m_rubberbandPen.SetWidth(1);
    m_RubberBand.SetPen(m_rubberbandPen);

    m_rubberbandorigin = wxPoint(INVALID_CELL,INVALID_CELL);

    wxColour axesColour(200, 245, 220);
    m_axesPen = wxPen(axesColour);
    m_axesBrush = wxBrush(axesColour);

    UpdateCanvasSize();

    SetDrawingMode(DRAWMODE_PENCIL);

    SetScrollRate(5, 5);
    Scroll(0,10000);
}

FiberedFractalTilerCanvas::~FiberedFractalTilerCanvas()
{
}

void FiberedFractalTilerCanvas::SetDrawingMode(int mode)
{
  m_pFrame->SetDrawingMode(mode);
  ResetRubberband();
  if (m_pFrame->GetDrawingMode() == DRAWMODE_PENCIL) {
    SetCursor(wxCURSOR_PENCIL);
  } else if (m_pFrame->GetDrawingMode() == DRAWMODE_RECTANGLE) {
    SetCursor(wxCURSOR_CROSS);
  } else if (m_pFrame->GetDrawingMode() == DRAWMODE_CLIPBOARD) {
    SetCursor(wxCURSOR_ARROW);
  }
}

//  Conversion between cell and screen coordinates
wxInt32 FiberedFractalTilerCanvas::XToCell(wxCoord x)
{
  int startX, startY;
  GetViewStart(&startX, &startY);
  int xUnit, yUnit;
  GetScrollPixelsPerUnit(&xUnit, &yUnit);

  wxInt32 retVal = ((x+(startX*xUnit)) / m_cellsize);
  if ((retVal < 0) || (retVal >= (int)m_pCellSet->GetDimX())) {
    retVal = INVALID_CELL;
  }
  int quad = m_pFrame->GetQuadrant();
  if ((quad == QUAD3) || (quad == QUAD4)) {
    retVal = -(int)(m_pCellSet->GetDimX() - retVal - 1);
  }
  return retVal;
}

wxInt32 FiberedFractalTilerCanvas::YToCell(wxCoord y)
{
  int startX, startY;
  GetViewStart(&startX, &startY);
  int xUnit, yUnit;
  GetScrollPixelsPerUnit(&xUnit, &yUnit);

  wxInt32 retVal = m_pCellSet->GetDimY() - ((y+(startY*yUnit)) / m_cellsize) - 1;
  if ((retVal < 0) || (retVal >= (int)m_pCellSet->GetDimY())) {
    retVal = INVALID_CELL;
  }
  int quad = m_pFrame->GetQuadrant();
  if ((quad == QUAD2) || (quad == QUAD3)) {
    retVal = -(int)(m_pCellSet->GetDimY() - retVal - 1);
  }
  return retVal;
}

wxCoord FiberedFractalTilerCanvas::CellToX(wxInt32 i)
{
  int startX, startY;
  GetViewStart(&startX, &startY);
  int xUnit, yUnit;
  GetScrollPixelsPerUnit(&xUnit, &yUnit);

  return ((i - m_pCellSet->m_minX) * m_cellsize) + (startX*xUnit);
}

wxCoord FiberedFractalTilerCanvas::CellToY(wxInt32 j)
{
  int startX, startY;
  GetViewStart(&startX, &startY);
  int xUnit, yUnit;
  GetScrollPixelsPerUnit(&xUnit, &yUnit);

  return ((m_pCellSet->GetDimY() - ((j - m_pCellSet->m_minY) + 1)) * m_cellsize) + (startY*yUnit);
}

// set the cell size and refresh display
void FiberedFractalTilerCanvas::SetCellSize(int cellsize)
{
  m_cellsize = cellsize;
  UpdateCanvasSize();
}

void FiberedFractalTilerCanvas::UpdateCanvasSize()
{
  int cw, ch;
  GetClientSize(&cw, &ch);

  int w = m_cellsize * m_pCellSet->GetDimX();
  int h = m_cellsize * m_pCellSet->GetDimY();

  SetVirtualSize((cw > w ? cw : w), (ch > h ? ch : h));

  Refresh(false);
}

// draw a cell
void FiberedFractalTilerCanvas::DrawCell(wxInt32 i, wxInt32 j, bool alive)
{
    wxClientDC dc(this);
    PrepareDC(dc);

    bool bAxis = (m_pFrame->GetMode() == MODE_SHAPE) ? false : (((i==0) || (j==0)) ? true : false);

    dc.SetPen(alive? m_pFrame->GetPenCellOn() : (bAxis ? m_axesPen : *wxWHITE_PEN));
    dc.SetBrush(alive? m_pFrame->GetBrushCellOn() : (bAxis ? m_axesBrush : *wxWHITE_BRUSH));

    DrawCell(i, j, dc);
}

void FiberedFractalTilerCanvas::DrawCell(wxInt32 i, wxInt32 j, wxDC &dc)
{
    wxCoord x = dc.LogicalToDeviceX(CellToX(i));
    wxCoord y = dc.LogicalToDeviceY(CellToY(j));

    // if cellsize is 1 or 2, there will be no grid
    switch (m_cellsize)
    {
        case 1:
            dc.DrawPoint(x, y);
            break;
        case 2:
            dc.DrawRectangle(x, y, 2, 2);
            break;
        default:
            dc.DrawRectangle(x + 1, y + 1, m_cellsize - 1, m_cellsize - 1);
    }
}

void FiberedFractalTilerCanvas::OnScroll(wxScrollWinEvent& event)
{
  event.Skip();
}

// event handlers
void FiberedFractalTilerCanvas::OnPaint(wxPaintEvent& WXUNUSED(event))
{
  wxPaintDC dc(this);
  DoPrepareDC(dc);

  wxRect  rect = wxRect(0, 0, m_cellsize*m_pCellSet->GetDimX(), m_cellsize*m_pCellSet->GetDimY());//GetUpdateRegion().GetBox();
  wxCoord x, y, w, h;

  // find damaged area
  x = rect.GetX();
  y = rect.GetY();
  w = rect.GetWidth();
  h = rect.GetHeight();

  //ClearBackground();  //  This causes problems due to incomplete redraws on scrolling
  dc.SetBackground(*wxLIGHT_GREY);
  dc.Clear();

  //  This is a super hack to make sure that the entire background is cleared because
  //  for some reason the edges weren't and were remaining transparent
  dc.SetBrush(*wxLIGHT_GREY_BRUSH);
  dc.DrawRectangle(0, 0, w+10000, h+10000);

  dc.SetBrush(*wxWHITE_BRUSH);
  dc.DrawRectangle(x, y, w, h);

  if (m_cellsize > 2) {
    dc.SetPen(*wxLIGHT_GREY_PEN);
    for (wxInt32 yy = y; yy <= (y + h - m_cellsize); yy += m_cellsize) {
      dc.DrawRectangle(x, yy, w, m_cellsize + 1);
    }
    for (wxInt32 xx = x; xx <= (x + w - m_cellsize); xx += m_cellsize) {
      dc.DrawLine(xx, y, xx, y + h);
    }
  }

  dc.SetPen(m_pFrame->GetPenCellOn());
  dc.SetBrush(m_pFrame->GetBrushCellOn());

  for (size_t m = 0; m < m_pCellSet->size(); m++) {
    if (m_pCellSet->at(m).m_bOn) {
      DrawCell(m_pCellSet->GetXFromIndex(m), m_pCellSet->GetYFromIndex(m), dc);
    }
  }

  if (m_pFrame->GetMode() == MODE_FRACTAL) {
    dc.SetPen(m_axesPen);
    dc.SetBrush(m_axesBrush);

    for (wxInt32 m=m_pCellSet->m_minY; m<=m_pCellSet->m_maxY; m++) {
      if (!m_pCellSet->GetCellValue(0, m)) {
        DrawCell(0, m, dc);
      }
    }
    for (wxInt32 m=m_pCellSet->m_minX; m<=m_pCellSet->m_maxX; m++) {
      if (!m_pCellSet->GetCellValue(m, 0)) {
        DrawCell(m, 0, dc);
      }
    }
  }
}

void FiberedFractalTilerCanvas::OnMouse(wxMouseEvent& event)
{
  if (event.Leaving()) {
    ((FiberedFractalTilerFrame*)wxGetApp().GetTopWindow())->SetStatusText(_(""));
    ResetRubberband();
    Refresh();
    return;
  }

  if (m_pFrame->GetDrawingMode() == DRAWMODE_CLIPBOARD) {
    if (event.RightIsDown()) {
      wxMenu* pMenu = new wxMenu();
      pMenu->Append(wxID_CUT, "Cut");
      pMenu->Append(wxID_COPY, "Copy");
      pMenu->Append(wxID_PASTE, "Paste");
      PopupMenu(pMenu);
      delete pMenu;
      return;
    }
  }

    // which cell are we pointing at?
    wxInt32 i = XToCell( event.GetX() );
    wxInt32 j = YToCell( event.GetY() );

    // set statusbar text
    if ((i == INVALID_CELL) || (j == INVALID_CELL)) {
      ((FiberedFractalTilerFrame *) wxGetApp().GetTopWindow())->SetStatusText(_(""));
      if (!((event.LeftUp() || event.RightUp()) && ((m_pFrame->GetDrawingMode() == DRAWMODE_RECTANGLE) || (m_pFrame->GetDrawingMode() == DRAWMODE_CLIPBOARD)))) {
        return;
      }
    } else {
      wxString msg;
      msg.Printf(_("Cell: (%d, %d)"), i, j);
      ((FiberedFractalTilerFrame *) wxGetApp().GetTopWindow())->SetStatusText(msg);
    }

    if ((event.LeftUp() || event.RightUp()) && (m_pFrame->GetDrawingMode() == DRAWMODE_RECTANGLE)) {
      wxRect emptyRect(0,0,0,0);
      wxRect rect = m_RubberBand.GetRect();

      if (rect == emptyRect) {
        return;
      }

      ResetRubberband();
      Refresh();
      wxInt32 x = XToCell(rect.GetX());
      wxInt32 y = YToCell(rect.GetY());

      bool alive = event.LeftUp();
      wxClientDC dc(this);
      PrepareDC(dc);
      dc.SetPen(alive? m_pFrame->GetPenCellOn() : *wxWHITE_PEN);
      dc.SetBrush(alive? m_pFrame->GetBrushCellOn() : *wxWHITE_BRUSH);

      int xnum = rect.GetWidth() / m_cellsize;
      int ynum = rect.GetHeight() / m_cellsize;
      for (int i=x; i<=(x+xnum); i++) {
        for (int j=(y-ynum); j<=y; j++) {
          m_pCellSet->SetCellValue(i, j, alive);
          DrawCell(i, j, dc);
        }
      }
      return;
    }

    // NOTE that wxMouseEvent::LeftDown() and wxMouseEvent::LeftIsDown()
    // have different semantics. The first one is used to signal that the
    // button was just pressed (i.e., in "button down" events); the second
    // one just describes the current status of the button, independently
    // of the mouse event type. LeftIsDown is typically used in "mouse
    // move" events, to test if the button is _still_ pressed.

    // is the button down?
    if (!event.LeftIsDown() && !event.RightIsDown())
    {
        m_status = MOUSE_NOACTION;
        return;
    }

    // was it pressed just now?
    if (event.LeftDown() || event.RightDown())
    {

      if (m_pFrame->GetDrawingMode() == DRAWMODE_RECTANGLE) {
        m_rubberbandorigin = event.GetPosition();
        m_mi = i;
        m_mj = j;
        return;
      } else {
        // yes: start a new action and toggle this cell
        m_status = (event.RightIsDown() ? MOUSE_ERASING : MOUSE_DRAWING);

        m_mi = i;
        m_mj = j;
        m_pCellSet->SetCellValue(i, j, m_status == MOUSE_DRAWING);
        DrawCell(i, j, m_status == MOUSE_DRAWING);
      }
    }
    else if (((m_mi != i) || (m_mj != j)) && (m_mi != INVALID_CELL) && (m_mj != INVALID_CELL))
    {
      if (m_pFrame->GetDrawingMode() == DRAWMODE_RECTANGLE) {
        wxPoint myActualPosition = event.GetPosition();
        int xIn, yIn, wIn, hIn;
        if (myActualPosition.x < m_rubberbandorigin.x) {
          xIn = myActualPosition.x;
          wIn = m_rubberbandorigin.x - myActualPosition.x;
        } else {
          xIn = m_rubberbandorigin.x;
          wIn = myActualPosition.x - m_rubberbandorigin.x;
        }
        if (myActualPosition.y < m_rubberbandorigin.y) {
          yIn = myActualPosition.y;
          hIn = m_rubberbandorigin.y - myActualPosition.y;
        } else {
          yIn = m_rubberbandorigin.y;
          hIn = myActualPosition.y - m_rubberbandorigin.y;
        }
        m_RubberBand.SetGeometry(xIn, yIn, wIn, hIn);
        return;
      } else if (m_status != MOUSE_NOACTION) {

        // no: continue ongoing action
        bool alive = (m_status == MOUSE_DRAWING);

        wxClientDC dc(this);
        PrepareDC(dc);

        // draw a line of cells using Bresenham's algorithm
        wxInt32 d, ii, jj, di, ai, si, dj, aj, sj;
        di = i - m_mi;
        ai = abs(di) << 1;
        si = (di < 0)? -1 : 1;
        dj = j - m_mj;
        aj = abs(dj) << 1;
        sj = (dj < 0)? -1 : 1;

        ii = m_mi;
        jj = m_mj;

        if (ai > aj)
        {
            // iterate over i
            d = aj - (ai >> 1);

            while (ii != i)
            {
                m_pCellSet->SetCellValue(ii, jj, alive);
                bool bAxis = (m_pFrame->GetMode() == MODE_SHAPE) ? false : (((ii==0) || (jj==0)) ? true : false);
                dc.SetPen(alive? m_pFrame->GetPenCellOn() : bAxis? m_axesPen : *wxWHITE_PEN);
                dc.SetBrush(alive? m_pFrame->GetBrushCellOn() : bAxis? m_axesBrush : *wxWHITE_BRUSH);
                DrawCell(ii, jj, dc);
                if (d >= 0)
                {
                    jj += sj;
                    d  -= ai;
                }
                ii += si;
                d  += aj;
            }
        }
        else
        {
            // iterate over j
            d = ai - (aj >> 1);

            while (jj != j)
            {
                m_pCellSet->SetCellValue(ii, jj, alive);
                bool bAxis = (m_pFrame->GetMode() == MODE_SHAPE) ? false : (((ii==0) || (jj==0)) ? true : false);
                dc.SetPen(alive? m_pFrame->GetPenCellOn() : bAxis? m_axesPen : *wxWHITE_PEN);
                dc.SetBrush(alive? m_pFrame->GetBrushCellOn() : bAxis? m_axesBrush : *wxWHITE_BRUSH);
                DrawCell(ii, jj, dc);
                if (d >= 0)
                {
                    ii += si;
                    d  -= aj;
                }
                jj += sj;
                d  += ai;
            }
        }

        // last cell
        m_pCellSet->SetCellValue(ii, jj, alive);
        bool bAxis = (m_pFrame->GetMode() == MODE_SHAPE) ? false : (((ii==0) || (jj==0)) ? true : false);
        dc.SetPen(alive? m_pFrame->GetPenCellOn() : bAxis? m_axesPen : *wxWHITE_PEN);
        dc.SetBrush(alive? m_pFrame->GetBrushCellOn() : bAxis? m_axesBrush : *wxWHITE_BRUSH);
        DrawCell(ii, jj, dc);
        m_mi = ii;
        m_mj = jj;
      }
    }
}

void FiberedFractalTilerCanvas::ResetRubberband()
{
  m_RubberBand.ClearOldRubberRect();
  m_rubberbandorigin = wxPoint(INVALID_CELL,INVALID_CELL);
}

void FiberedFractalTilerCanvas::OnSize(wxSizeEvent& event)
{
  // allow default processing
  event.Skip();
}

void FiberedFractalTilerCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
  // do nothing. I just don't want the background to be erased, you know.
}

void FiberedFractalTilerCanvas::OnCut(wxCommandEvent& WXUNUSED(event))
{
  //PutTileTypesOnClipboard(m_selectedList, true);
  //m_selectedList.clear();
  Refresh();
}

void FiberedFractalTilerCanvas::OnCopy(wxCommandEvent& WXUNUSED(event))
{
  //PutTileTypesOnClipboard(m_selectedList, false);
}

void FiberedFractalTilerCanvas::OnPaste(wxCommandEvent& WXUNUSED(event))
{

}