/////////////////////////////////////////////////////////////////////////////
// Name:        FiberedFractalTiler.h
// Purpose:     Generate TAM tilesets which self-assemble into designed shapes
// Author:      Matthew Patitz <mpatitz@yahoo.com>
// Modified by:
// Created:     Nov/2007
// Copyright:   (c) 2008, Matthew John Patitz
// Licence:     GPL
/////////////////////////////////////////////////////////////////////////////

#ifndef _FiberedFractalTiler_APP_H_
#define _FiberedFractalTiler_APP_H_

// for compilers that support precompilation, includes "wx/wx.h"
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

// for all others, include the necessary headers
#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include <wx/filename.h>
#include <wx/html/helpctrl.h>
#include <wx/choice.h>
#include <vector>
#include "points.h"
#include "tiles.h"
#include "wxRubberBand.h"


#define CELLSET_DEFAULT_MAX 100

#define MODE_SHAPE    0
#define MODE_FRACTAL  1

#define DRAWMODE_PENCIL     0
#define DRAWMODE_RECTANGLE  1
#define DRAWMODE_CLIPBOARD  2

#define FILLER_LABEL  "."

const int ID_CTRL_SCALE     = 500;
const int ID_CTRL_TSET      = 501;
const int ID_CTRL_SEED      = 502;
const int ID_BUTT_GEN       = 503;
const int ID_SETGRIDSIZE    = 504;
const int ID_RADIO_MODE     = 506;
const int ID_CANVAS_WIN     = 507;
const int ID_CTRL_MSGS      = 508;
const int ID_CTRL_QUAD      = 509;
const int ID_TILESET_BROWSE = 510;
const int ID_SEED_BROWSE    = 511;
const int ID_STAGE          = 512;
const int ID_BUTT_FORWARD   = 513;
const int ID_BUTT_RESET     = 514;

const int INVALID_CELL    = -9999999;

class Cell
{
public:
  Cell() { m_bOn = false; };
  Cell(const Cell& cIn) { m_bOn = cIn.m_bOn; };
  
  void Flip() { m_bOn = !m_bOn; };

  bool m_bOn;
};


class CellSet : public std::vector<Cell>
{
public:
  CellSet();
  CellSet(wxInt32 minX, wxInt32 maxX, wxInt32 minY, wxInt32 maxY);
  ~CellSet();

  void SetSize(wxInt32 minX, wxInt32 maxX, wxInt32 minY, wxInt32 maxY);

  void SetCellValue(wxInt32 x, wxInt32 y, bool bVal);
  bool GetCellValue(wxInt32 x, wxInt32 y);

  void Clear();

  size_t GetIndex(wxInt32 x, wxInt32 y);
  wxInt32 GetXFromIndex(size_t ind);
  wxInt32 GetYFromIndex(size_t ind);

  size_t GetDimX();
  size_t GetDimY();

  wxInt32 m_minX;
  wxInt32 m_maxX;
  wxInt32 m_minY;
  wxInt32 m_maxY;
};

class FiberedFractalTilerFrame;
// --------------------------------------------------------------------------
// FiberedFractalTilerCanvas
// --------------------------------------------------------------------------

// Note that in FiberedFractalTilerCanvas, all cell coordinates are
// named i, j, while screen coordinates are named x, y.

class FiberedFractalTilerCanvas : public wxScrolledWindow
{
public:
    // ctor and dtor
    FiberedFractalTilerCanvas(wxWindow* parent, FiberedFractalTilerFrame* pFrame, CellSet* pCellSet);
    virtual ~FiberedFractalTilerCanvas();

    // view management
    int  GetCellSize() const { return m_cellsize; };
    void SetCellSize(int cellsize);
    void UpdateCanvasSize();

    // drawing
    void DrawCell(wxInt32 i, wxInt32 j, bool alive);
    void ResetRubberband();
    void SetDrawingMode(int mode);

private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_CLASS(FiberedFractalTilerCanvas)
    DECLARE_EVENT_TABLE()

    // draw a cell (parametrized by DC)
    void DrawCell(wxInt32 i, wxInt32 j, wxDC &dc);

    // event handlers
    void OnPaint(wxPaintEvent& event);
    void OnScroll(wxScrollWinEvent& event);
    void OnMouse(wxMouseEvent& event);
    void OnSize(wxSizeEvent& event);
    void OnEraseBackground(wxEraseEvent& event);
    void OnCut(wxCommandEvent& event);
    void OnCopy(wxCommandEvent& event);
    void OnPaste(wxCommandEvent& event);

    // conversion between cell and screen coordinates
    wxInt32 XToCell(wxCoord x);
    wxInt32 YToCell(wxCoord y);
    wxCoord CellToX(wxInt32 i);
    wxCoord CellToY(wxInt32 j);

    // what is the user doing?
    enum MouseStatus
    {
        MOUSE_NOACTION,
        MOUSE_DRAWING,
        MOUSE_ERASING
    };

    FiberedFractalTilerFrame*    m_pFrame;
    CellSet*      m_pCellSet;
    int           m_cellsize;        // current cell size, in pixels
    MouseStatus   m_status;          // what is the user doing?
    wxInt32       m_mi, m_mj;        // last mouse position

    wxRubberBand  m_RubberBand;
    wxPoint       m_rubberbandorigin;
    wxPen         m_rubberbandPen;

    wxPen         m_axesPen;
    wxBrush       m_axesBrush;
};


// --------------------------------------------------------------------------
// FiberedFractalTilerFrame
// --------------------------------------------------------------------------

class FiberedFractalTilerFrame : public wxFrame
{
public:
    // ctor and dtor
    FiberedFractalTilerFrame();
    virtual ~FiberedFractalTilerFrame();

    // member functions
    void UpdateInfoText();
    void UpdateUI();

    int GetMode() { return m_mode; };
    int GetDrawingMode() { return m_drawMode; };
    void SetDrawingMode(int mode) { m_drawMode = mode; };
    int GetQuadrant();

    wxPen& GetPenCellOn() { return m_penCellOn; };
    wxBrush& GetBrushCellOn() { return m_brushCellOn; };

    wxHtmlHelpController    m_helpController;
private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()

    // event handlers
    void OnMenu(wxCommandEvent& event);
    void OnOpen(wxCommandEvent& event);
    void OnSave(wxCommandEvent& event);
    void OnSaveAs(wxCommandEvent& event);
    void OnZoom(wxCommandEvent& event);
    void OnGridSize(wxCommandEvent& event);
    void OnColor(wxCommandEvent& event);
    void OnScale(wxCommandEvent& event);
    void OnMode(wxCommandEvent& event);
    void OnForward(wxCommandEvent& event);
    void OnReset(wxCommandEvent& event);
    void OnQuadrant(wxCommandEvent& event);
    void OnTilesetFile(wxCommandEvent& event);
    void OnSeedFile(wxCommandEvent& event);
    void OnBrowse(wxCommandEvent& event);
    void OnGenerate(wxCommandEvent& event);
    void OnClose(wxCloseEvent& event);

    void AddMessage(wxString sMsg, bool bClear = false);

    bool GenerateTilesForShape(PointSet* pPS, TileSet* pTS, long scale);

    FiberedFractalTilerCanvas *m_canvas;

    CellSet         m_cellSet;
    CellSet         m_cellSetStage1;
    PointSet        m_pointSet;
    long            m_scale;
    int             m_mode;
    int             m_drawMode;
    int             m_stage;
    int             m_quadrant;
    wxInt32         m_maxXOn;
    wxInt32         m_maxYOn;
    wxFileName      m_fnShapeDefFile;
    wxFileName      m_fnTilesetFile;
    wxFileName      m_fnSeedFile;
    wxTextCtrl     *m_ptcStage;
    wxTextCtrl     *m_ptcScale;
    wxTextCtrl     *m_ptcTilesetFile;
    wxTextCtrl     *m_ptcSeedFile;
    wxTextCtrl     *m_ptcMessageBox;
    wxChoice       *m_pccQuadrant;
    wxRadioBox     *m_pradioMode;
    wxButton       *m_pbuttonForward;
    wxButton       *m_pbuttonReset;
    wxButton       *m_pbuttonGenerate;
    wxColour        m_colCellOn;
    wxPen           m_penCellOn;
    wxBrush         m_brushCellOn;

    wxStreamToTextRedirector *m_pStreamRedirect;
};


// --------------------------------------------------------------------------
// FiberedFractalTilerApp
// --------------------------------------------------------------------------

class FiberedFractalTilerApp : public wxApp
{
public:
    virtual bool OnInit();
};

#endif  // _FiberedFractalTiler_APP_H_
